"""An advanced template tag for caching in django: versioning, compress, partial caching, easy inheritance"""

import pkg_resources
from os import path
from setuptools.config import read_configuration


def _extract_version(package_name):
    try:
        # if package is installed
        version = pkg_resources.get_distribution(package_name).version
    except pkg_resources.DistributionNotFound:
        # if not installed, so we must be in source, with ``setup.cfg`` available
        _conf = read_configuration(path.join(
            path.dirname(__file__), '..', 'setup.cfg')
        )
        version = _conf['metadata']['version']

    return version


EXACT_VERSION = _extract_version('erudit_django_adv_cache_tag')
VERSION = tuple(int(part) for part in EXACT_VERSION.split('.') if str(part).isnumeric())



import hashlib
import pickle
import time
import zlib

from copy import deepcopy
from datetime import datetime

from django.conf import settings
from django.utils.encoding import force_bytes
from django.utils.safestring import SafeText

from django import VERSION as django_version
from django.test.utils import override_settings
from django.utils.http import urlquote

from adv_cache_tag.compat import get_cache, template
from adv_cache_tag.tag import CacheTag

from .compat import TestCase


# Force some settings to not depend on the external ones
@override_settings(

    DEBUG = False,
    TEMPLATE_DEBUG = False,

    # Force using memory cache
    CACHES = {
        'default': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'LOCATION': 'default-cache',
        },
        'foo': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            'LOCATION': 'foo-cache',
        },
    },

    # Used to compose RAW tags
    SECRET_KEY = 'm-92)2et+&&m5f&#jld7-_1qanq*n9!z90xc@+wx6y8d6y#w6t',

    # Reset default config
    ADV_CACHE_VERSIONING = False,
    ADV_CACHE_COMPRESS = False,
    ADV_CACHE_COMPRESS_LEVEL = zlib.Z_DEFAULT_COMPRESSION,
    ADV_CACHE_COMPRESS_SPACES= False,
    ADV_CACHE_INCLUDE_PK = False,
    ADV_CACHE_BACKEND = 'default',
    ADV_CACHE_VERSION = '',
    ADV_CACHE_RESOLVE_NAME = False,

    # For django >= 1.8 (RemovedInDjango110Warning appears in 1.9)
    TEMPLATES = [
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'APP_DIRS': True,
            'OPTIONS': {
                'debug': False
            }
        },
    ]
)
class BasicTestCase(TestCase):
    """First basic test case to be able to test python/django compatibility."""

    @classmethod
    def reload_config(cls):
        """Resest the ``CacheTag`` configuration from current settings"""
        CacheTag.options.versioning = getattr(settings, 'ADV_CACHE_VERSIONING', False)
        CacheTag.options.compress = getattr(settings, 'ADV_CACHE_COMPRESS', False)
        CacheTag.options.compress_level = getattr(settings, 'ADV_CACHE_COMPRESS_LEVEL', False)
        CacheTag.options.compress_spaces = getattr(settings, 'ADV_CACHE_COMPRESS_SPACES', False)
        CacheTag.options.include_pk = getattr(settings, 'ADV_CACHE_INCLUDE_PK', False)
        CacheTag.options.cache_backend = getattr(settings, 'ADV_CACHE_BACKEND', 'default')
        CacheTag.options.resolve_fragment = getattr(settings, 'ADV_CACHE_RESOLVE_NAME', False)

        # generate a token for this site, based on the secret_key
        CacheTag.RAW_TOKEN = 'RAW_' + hashlib.sha1(
            b'RAW_TOKEN_SALT1' + force_bytes(hashlib.sha1(
                b'RAW_TOKEN_SALT2' + force_bytes(settings.SECRET_KEY)
            ).hexdigest())
        ).hexdigest()

        # tokens to use around the already parsed parts of the cached template
        CacheTag.RAW_TOKEN_START = template.BLOCK_TAG_START + CacheTag.RAW_TOKEN + \
                                   template.BLOCK_TAG_END
        CacheTag.RAW_TOKEN_END = template.BLOCK_TAG_START + 'end' + CacheTag.RAW_TOKEN + \
                                 template.BLOCK_TAG_END

    def setUp(self):
        """Clean stuff and create an object to use in templates, and some counters."""
        super(BasicTestCase, self).setUp()

        # Clear the cache
        for cache_name in settings.CACHES:
            get_cache(cache_name).clear()

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        # And an object to cache in template
        self.obj = {
            'pk': 42,
            'name': 'foobar',
            'get_name': self.get_name,
            'get_foo': self.get_foo,
            'updated_at': datetime(2015, 10, 27, 0, 0, 0),
        }

        # To count the number of calls of ``get_name`` and ``get_foo``.
        self.get_name_called = 0
        self.get_foo_called = 0

    def get_name(self):
        """Called in template when asking for ``obj.get_name``."""
        self.get_name_called += 1
        return self.obj['name']

    def get_foo(self):
        """Called in template when asking for ``obj.get_foo``."""
        self.get_foo_called += 1
        return 'foo %d' % self.get_foo_called

    def tearDown(self):
        """Clear caches at the end."""

        for cache_name in settings.CACHES:
            get_cache(cache_name).clear()

        super(BasicTestCase, self).tearDown()

    @classmethod
    def tearDownClass(cls):
        """At the very end of all theses tests, we reload the CacheTag config."""

        # Reset CacheTag config after the end of ``override_settings``
        cls.reload_config()

        super(BasicTestCase, cls).tearDownClass()

    @staticmethod
    def get_template_key(fragment_name, vary_on=None, prefix='template.cache'):
        """Compose the cache key of a template."""
        if vary_on is None:
            vary_on = ()
        key = ':'.join([urlquote(force_bytes(var)) for var in vary_on])
        args = hashlib.md5(force_bytes(key))
        return (prefix + '.%s.%s') % (fragment_name, args.hexdigest())

    def render(self, template_text, extend_context_dict=None):
        """Utils to render a template text with a context given as a dict."""
        context_dict = {'obj': self.obj}
        if extend_context_dict:
            context_dict.update(extend_context_dict)
        return template.Template(template_text).render(template.Context(context_dict))

    def assertStripEqual(self, first, second):
        """Like ``assertEqual`` for strings, but after calling ``strip`` on both arguments."""
        if first:
            first = first.strip()
        if second:
            second = second.strip()

        self.assertEqual(first, second)

    def assertNotStripEqual(self, first, second):
        """Like ``assertNotEqual`` for strings, but after calling ``strip`` on both arguments."""
        if first:
            first = first.strip()
        if second:
            second = second.strip()

        self.assertNotEqual(first, second)

    def test_default_cache(self):
        """This test is only to validate the testing procedure."""

        expected = "foobar"

        t = """
            {% load cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        self.assertStripEqual(get_cache('default').get(key), expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    def test_adv_cache(self):
        """Test default behaviour with default settings."""

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # But it should NOT be the exact content as adv_cache_tag adds a version
        self.assertNotStripEqual(get_cache('default').get(key), expected)

        # It should be the version from `adv_cache_tag`
        cache_expected = b"1::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    def test_timeout_value(self):
        "Test that timeout value is ``None`` or an integer."""

        ok_values = ('0', '1', '9999', '"0"', '"1"', '"9999"', 'None')
        ko_values = ('-1', '-9999', '"-1"', '"-9999"', '"foo"', '""', '12.3', '"12.3"')

        t = """
            {%% load adv_cache %%}
            {%% cache %s test_cached_template obj.pk obj.updated_at %%}
                {{ obj.get_name }}
            {%% endcache %%}
        """

        for value in ok_values:
            def test_value(value):
                self.render(t % value)

            if hasattr(self, 'subTest'):
                with self.subTest(value=value):
                    test_value(value)
            else:
                test_value(value)

        for value in ko_values:
            def test_value(value):
                with self.assertRaises(template.TemplateSyntaxError) as raise_context:
                    self.render(t % value)
                self.assertIn(
                    'tag got a non-integer (or None) timeout value',
                    str(raise_context.exception)
                )

            if hasattr(self, 'subTest'):
                with self.subTest(value=value):
                    test_value(value)
            else:
                test_value(value)

    def test_quoted_fragment_name(self):
        """Test quotes behaviour around the fragment name."""

        t = """
            {% load adv_cache %}
            {% cache 1 "test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        with self.assertRaises(ValueError) as raise_context:
            self.render(t)
        self.assertIn('incoherent', str(raise_context.exception))

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template" obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        with self.assertRaises(ValueError) as raise_context:
            self.render(t)
        self.assertIn('incoherent', str(raise_context.exception))

        t = """
            {% load adv_cache %}
            {% cache 1 'test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        with self.assertRaises(ValueError) as raise_context:
            self.render(t)
        self.assertIn('incoherent', str(raise_context.exception))

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template" obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        with self.assertRaises(ValueError) as raise_context:
            self.render(t)
        self.assertIn('incoherent', str(raise_context.exception))

        t = """
            {% load adv_cache %}
            {% cache 1 "test_cached_template" obj.pk "foo" obj.updated_at %}
                {{ obj.get_name }} foo
            {% endcache %}
        """
        expected = "foobar foo"
        self.assertStripEqual(self.render(t), expected)
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], 'foo', self.obj['updated_at']])
        self.assertEqual(  # no quotes arround `test_cached_template`
            key, 'template.cache.test_cached_template.f2f294788f4c38512d3b544ce07befd0')
        cache_expected = b"1::\n                foobar foo"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        t = """
            {% load adv_cache %}
            {% cache 1 'test_cached_template' obj.pk "bar" obj.updated_at %}
                {{ obj.get_name }} bar
            {% endcache %}
        """
        expected = "foobar bar"
        self.assertStripEqual(self.render(t), expected)
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], 'bar', self.obj['updated_at']])
        self.assertEqual(  # no quotes arround `test_cached_template`
            key, 'template.cache.test_cached_template.8bccdefc91dc857fc02f6938bf69b816')
        cache_expected = b"1::\n                foobar bar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

    @override_settings(
        ADV_CACHE_VERSIONING = True,
    )
    def test_versioning(self):
        """Test with ``ADV_CACHE_VERSIONING`` set to ``True``."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache

        # ``obj.updated_at`` is not in the key anymore, serving as the object version
        key = self.get_template_key('test_cached_template', vary_on=[self.obj['pk']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.a1d0c6e83f027327d8461063f4ac58a6')

        # It should be in the cache, with the ``updated_at`` in the version
        cache_expected = b"1::2015-10-27 00:00:00::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

        # We can update the date
        self.obj['updated_at'] = datetime(2015, 10, 28, 0, 0, 0)

        # Render with the new date, we should miss the cache because of the new "version
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 2)  # One more

        # It should be in the cache, with the new ``updated_at`` in the version
        cache_expected = b"1::2015-10-28 00:00:00::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 2)  # Still 2

    @override_settings(
        ADV_CACHE_INCLUDE_PK = True,
    )
    def test_primary_key(self):
        """Test with ``ADV_CACHE_INCLUDE_PK`` set to ``True``."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache

        # We add the pk as a part to the fragment name
        key = self.get_template_key('test_cached_template.%s' % self.obj['pk'],
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.42.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache
        cache_expected = b"1::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    @override_settings(
        ADV_CACHE_COMPRESS_SPACES = True,
    )
    def test_space_compression(self):
        """Test with ``ADV_CACHE_COMPRESS_SPACES`` set to ``True``."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache, with only one space instead of many white spaces
        cache_expected = b"1:: foobar "
        # Test with ``assertEqual``, not ``assertStripEqual``
        self.assertEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    @override_settings(
        ADV_CACHE_COMPRESS = True,
    )
    def test_compression(self):
        """Test with ``ADV_CACHE_COMPRESS`` set to ``True``."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        # We don't use new lines here because too complicated to set empty lines with only
        # spaces in a docstring with we'll have to compute the compressed version
        t = "{% load adv_cache %}{% cache 1 test_cached_template obj.pk obj.updated_at %}" \
            "  {{ obj.get_name }}  {% endcache %}"

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache, compressed
        # We use ``SafeText`` as django does in templates
        compressed = zlib.compress(pickle.dumps(SafeText("  foobar  ")), -1)
        cache_expected = b'1::' + compressed
        # Test with ``assertEqual``, not ``assertStripEqual``
        self.assertEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

        # Changing the compression level should not invalidate the cache
        CacheTag.options.compress_level = 9
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

        # But if the cache is invalidated, the new one will use this new level
        get_cache('default').delete(key)
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 2)  # One more
        compressed = zlib.compress(pickle.dumps(SafeText("  foobar  ")), 9)
        cache_expected = b'1::' + compressed
        self.assertEqual(get_cache('default').get(key), cache_expected)


    @override_settings(
        ADV_CACHE_COMPRESS = True,
        ADV_CACHE_COMPRESS_SPACES = True,
    )
    def test_full_compression(self):
        """Test with ``ADV_CACHE_COMPRESS`` and ``ADV_CACHE_COMPRESS_SPACES`` set to ``True``."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache, compressed
        # We DON'T use ``SafeText`` as in ``test_compression`` because with was converted back
        # to a real string when removing spaces
        compressed = zlib.compress(pickle.dumps(" foobar "))
        cache_expected = b'1::' + compressed
        # Test with ``assertEqual``, not ``assertStripEqual``
        self.assertEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    @override_settings(
        ADV_CACHE_BACKEND = 'foo',
    )
    def test_cache_backend(self):
        """Test with ``ADV_CACHE_BACKEND`` to another value than ``default``."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache
        cache_expected = b"1::\n                foobar"

        # But not in the ``default`` cache
        self.assertIsNone(get_cache('default').get(key))

        # But in the ``foo`` cache
        self.assertStripEqual(get_cache('foo').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    @override_settings(
        ADV_CACHE_COMPRESS_SPACES = True,
    )
    def test_partial_cache(self):
        """Test the ``nocache`` templatetag."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar  foo 1  !!"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
                {% nocache %}
                    {{ obj.get_foo }}
                {% endnocache %}
                !!
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)
        self.assertEqual(self.get_foo_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache, with the RAW part
        cache_expected = b"1:: foobar {%endRAW_38a11088962625eb8c913e791931e2bc2e3c7228%} " \
                         b"{{obj.get_foo}} {%RAW_38a11088962625eb8c913e791931e2bc2e3c7228%} !! "
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache but not for ``get_foo``
        expected = "foobar  foo 2  !!"
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1
        self.assertEqual(self.get_foo_called, 2)  # One more call to the non-cached part

    @override_settings(
        ADV_CACHE_VERSIONING = True,
    )
    def test_internal_version(self):
        """Test a cache with `Meta.internal_version` set."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache_test %}
            {% cache_with_version 1 test_cache_with_version obj.pk %}
                {{ obj.get_name }}
            {% endcache_with_version %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # It should be in the cache, with the ``internal_version`` in the version
        key = 'template.cache_with_version.test_cache_with_version.a1d0c6e83f027327d8461063f4ac58a6'
        cache_expected = b"1|v1::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        self.get_name_called = 0
        # Calling it a new time should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 0)

        # Changing the interval version should miss the cache
        from .testproject.adv_cache_test_app.templatetags.adv_cache_test import InternalVersionTag
        InternalVersionTag.options.internal_version = 'v2'
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # It should be in the cache, with the new ``internal_version`` in the version
        key = 'template.cache_with_version.test_cache_with_version.a1d0c6e83f027327d8461063f4ac58a6'
        cache_expected = b"1|v2::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

    def test_new_class(self):
        """Test a new class based on ``CacheTag``."""

        expected = "foobar  foo 1  !!"

        t = """
            {% load adv_cache_test %}
            {% cache_test 1 multiplicator test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
                {% nocache_test %}
                    {{ obj.get_foo }}
                {% endnocache_test %}
                !!
            {% endcache_test %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t, {'multiplicator': 10}), expected)
        self.assertEqual(self.get_name_called, 1)
        self.assertEqual(self.get_foo_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']],
                                    prefix='template.cache_test')
        self.assertEqual(
            key, 'template.cache_test.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache, with the RAW part
        cache_expected = b"1:: foobar {%endRAW_38a11088962625eb8c913e791931e2bc2e3c7228%} " \
                         b"{{obj.get_foo}} {%RAW_38a11088962625eb8c913e791931e2bc2e3c7228%} !! "
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # We'll check that our multiplicator was really applied
        cache = get_cache('default')
        expire_at = cache._expire_info[cache.make_key(key, version=None)]
        now = time.time()
        # In more that one second (default expiry we set) and less than ten
        self.assertTrue(now + 1 < expire_at < now + 10)

        # Render a second time, should hit the cache but not for ``get_foo``
        expected = "foobar  foo 2  !!"
        self.assertStripEqual(self.render(t, {'multiplicator': 10}), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1
        self.assertEqual(self.get_foo_called, 2)  # One more call to the non-cached part

    @override_settings(
        ADV_CACHE_RESOLVE_NAME = True,
    )
    def test_resolve_fragment_name(self):
        """Test passing the fragment name as a variable."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 fragment_name obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t, {'fragment_name': 'test_cached_template'}), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # But it should NOT be the exact content as adv_cache_tag adds a version
        self.assertNotStripEqual(get_cache('default').get(key), expected)

        # It should be the version from `adv_cache_tag`
        cache_expected = b"1::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t, {'fragment_name': 'test_cached_template'}), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

        # Using an undefined variable should fail
        t = """
            {% load adv_cache %}
            {% cache 1 undefined_fragment_name obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        with self.assertRaises(template.VariableDoesNotExist) as raise_context:
            self.render(t, {'fragment_name': 'test_cached_template'})
        self.assertIn('undefined_fragment_name', str(raise_context.exception))

    @override_settings(
        ADV_CACHE_RESOLVE_NAME = True,
    )
    def test_passing_fragment_name_as_string(self):
        """Test passing the fragment name as a variable."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 "test_cached_template" obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # But it should NOT be the exact content as adv_cache_tag adds a version
        self.assertNotStripEqual(get_cache('default').get(key), expected)

        # It should be the version from `adv_cache_tag`
        cache_expected = b"1::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    def test_using_argument(self):
        """Test passing the cache backend to use with the `using=` arg to the templatetag."""

        expected = "foobar"

        t = """
            {% load adv_cache %}
            {% cache 1 test_cached_template obj.pk obj.updated_at using=foo %}
                {{ obj.get_name }}
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']])
        self.assertEqual(
            key, 'template.cache.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache
        cache_expected = b"1::\n                foobar"

        # But not in the ``default`` cache
        self.assertIsNone(get_cache('default').get(key))

        # But in the ``foo`` cache
        self.assertStripEqual(get_cache('foo').get(key), cache_expected)

        # Render a second time, should hit the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1

    @override_settings(
        ADV_CACHE_COMPRESS_SPACES = True,
    )
    def test_loading_libraries_in_nocache(self):
        """Test that needed libraries are loaded in the nocache block."""

        # Reset CacheTag config with default value (from the ``override_settings``)
        self.reload_config()

        expected = "foobar FoOoO   FOO 1FOO 1 FoOoO  !!"

        t = """
            {% load adv_cache other_tags %}
            {% cache 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }} {% insert_foo %}
                {% nocache %}
                    {% load other_filters %}
                    {{ obj.get_foo|double_upper }} {% insert_foo %}
                {% endnocache %}
                !!
            {% endcache %}
        """

        # Render a first time, should miss the cache
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)
        self.assertEqual(self.get_foo_called, 1)

        # Render a second time, should hit the cache but not for ``get_foo``
        expected = "foobar FoOoO   FOO 2FOO 2 FoOoO  !!"
        self.assertStripEqual(self.render(t), expected)
        self.assertEqual(self.get_name_called, 1)  # Still 1
        self.assertEqual(self.get_foo_called, 2)  # One more call to the non-cached part

    def set_template_debug_true(self):
        if django_version < (1, 8):
            return override_settings(TEMPLATE_DEBUG=True)

        # not so simple now, it's an option of a template backend
        templates_settings_copy = deepcopy(settings.TEMPLATES)
        for template_settings in templates_settings_copy:
            if template_settings['BACKEND'] == 'django.template.backends.django.DjangoTemplates':
                template_settings.setdefault('OPTIONS', {})['debug'] = True
        return override_settings(TEMPLATES=templates_settings_copy)

    def test_failure_when_setting_cache(self):
        """Test that the template is correctly rendered even if the cache cannot be filled."""

        expected = "foobar"

        t = """
            {% load adv_cache_test %}
            {% cache_set_fail 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache_set_fail %}
        """

        # Render a first time, should still be rendered
        self.assertStripEqual(self.render(t), expected)

        # Now the rendered template should NOT be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']],
                                    prefix='template.cache_set_fail')
        self.assertEqual(
            key, 'template.cache_set_fail.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # But not in the ``default`` cache
        self.assertIsNone(get_cache('default').get(key))

        # It should raise if templates debug mode is activated
        with self.set_template_debug_true():
            with self.assertRaises(ValueError) as raise_context:
                self.render(t)
            self.assertIn('boom set', str(raise_context.exception))

    def test_failure_when_getting_cache(self):
        """Test that the template is correctly rendered even if the cache cannot be read."""

        expected = "foobar"

        t = """
            {% load adv_cache_test %}
            {% cache_get_fail 1 test_cached_template obj.pk obj.updated_at %}
                {{ obj.get_name }}
            {% endcache_get_fail %}
        """

        # Render a first time, should still be rendered
        self.assertStripEqual(self.render(t), expected)

        # Now the rendered template should be in cache
        key = self.get_template_key('test_cached_template',
                                    vary_on=[self.obj['pk'], self.obj['updated_at']],
                                    prefix='template.cache_get_fail')
        self.assertEqual(
            key, 'template.cache_get_fail.test_cached_template.0cac9a03d5330dd78ddc9a0c16f01403')

        # It should be in the cache
        cache_expected = b"1::\n                foobar"
        self.assertStripEqual(get_cache('default').get(key), cache_expected)

        # It should raise if templates debug mode is activated
        with self.set_template_debug_true():
            with self.assertRaises(ValueError) as raise_context:
                self.render(t)
            self.assertIn('boom get', str(raise_context.exception))



from django import template

from adv_cache_tag.tag import CacheTag, Node

register = template.Library()


class TestNode(Node):
    def __init__(self, nodename, nodelist, expire_time, multiplicator, fragment_name, vary_on):
        """ Save the multiplicator variable in the node (not resolved yet) """
        super(TestNode, self).__init__(nodename, nodelist, expire_time, fragment_name, vary_on)
        self.multiplicator = multiplicator


class TestCacheTag(CacheTag):
    class Meta(CacheTag.Meta):
        compress_spaces = True

    Node = TestNode

    @classmethod
    def get_template_node_arguments(cls, tokens):
        """ Check validity of tokens and return them as ready to be passed to the Node class """
        if len(tokens) < 4:
            raise template.TemplateSyntaxError(
                "'%r' tag requires at least 3 arguments." % tokens[0])
        return tokens[1], tokens[2], tokens[3], tokens[4:]

    def prepare_params(self):
        """ Resolve the multiplicator variable to it's real content """
        self.multiplicator = int(template.Variable(self.node.multiplicator).resolve(self.context))
        super(TestCacheTag, self).prepare_params()

    def get_expire_time(self):
        """Update the expiry time with the multiplicator."""
        expiry_time = super(TestCacheTag, self).get_expire_time()
        return self.multiplicator * expiry_time

TestCacheTag.register(register, 'cache_test', 'nocache_test')


class FailingCacheSetCacheTag(CacheTag):
    def cache_set(self, to_cache):
        raise ValueError('boom set')

FailingCacheSetCacheTag.register(register, 'cache_set_fail')


class FailingCacheGetCacheTag(CacheTag):
    def cache_get(self):
        raise ValueError('boom get')

FailingCacheGetCacheTag.register(register, 'cache_get_fail')


class InternalVersionTag(CacheTag):
    class Meta(CacheTag.Meta):
        internal_version = 'v1'


InternalVersionTag.register(register, 'cache_with_version')



# django-adv-cache-tag / Copyright Stephane "Twidi" Angel <s.angel@twidi.com> / MIT License

from django import template

from adv_cache_tag.tag import CacheTag

register = template.Library()

# Register the default class with the "cache" and "nocache" block names
CacheTag.register(register)



# django-adv-cache-tag / Copyright Stephane "Twidi" Angel <s.angel@twidi.com> / MIT License

import hashlib
import logging
import pickle
import re
import zlib

from urllib.parse import quote

from django import VERSION as django_version
from django.conf import settings
from django.utils.encoding import smart_str, force_bytes

from .compat import get_cache, get_template_libraries, template


try:
    template.TokenType  # django >= 2.1 only
except AttributeError:
    TOKEN_TEXT = template.TOKEN_TEXT
    TOKEN_VAR = template.TOKEN_VAR
    TOKEN_BLOCK = template.TOKEN_BLOCK
    TOKEN_COMMENT = template.TOKEN_COMMENT
else:
    TOKEN_TEXT = template.TokenType.TEXT
    TOKEN_VAR = template.TokenType.VAR
    TOKEN_BLOCK = template.TokenType.BLOCK
    TOKEN_COMMENT = template.TokenType.COMMENT


logger = logging.getLogger('adv_cache_tag')


def is_template_debug_activated():
    if django_version < (1, 8):
        return settings.TEMPLATE_DEBUG

    # not so simple now, it's an option of a template backend
    for template_settings in settings.TEMPLATES:
        if template_settings['BACKEND'] == 'django.template.backends.django.DjangoTemplates':
            return bool(template_settings.get('OPTIONS', {}).get('debug', False))

    return False


class Node(template.Node):
    """
    It's a normal template Node, with parameters defined in __init__ and rendering
    This class can be extended in your own main cache class, by redefining a `Node`
    class descending on this one.
    """

    def __init__(self, nodename, nodelist, expire_time, fragment_name, vary_on):
        """
        Define parameters to be used by the templatetag. The same as for
        the default `cache` templatetag in django, except for `nodename`
        which is the name used to call this templatetag ("cache" by default)

        If versioning is activated, the last argument in `vary_on` is popped
        and used for this purpose.

        If the `include_pk` option is activated, the first argument in `vary_on`
        will be used as the `pk` (but not removed from `vary_on`.
        """
        super(Node, self).__init__()
        self.nodename = nodename
        self.nodelist = nodelist
        self.expire_time = template.Variable(expire_time)
        self.fragment_name = template.Variable(fragment_name)

        self.cache_backend = None
        if vary_on and vary_on[-1].startswith('using='):
            self.cache_backend = vary_on.pop()[len('using='):]

        self.version = None
        if self._cachetag_class_.options.versioning:
            try:
                self.version = template.Variable(vary_on.pop())
            except Exception:
                self.version = None

        self.vary_on = vary_on

    def render(self, context):
        """
        Render the template by calling the render method of the main
        cache object.
        """
        return self._cachetag_class_(self, context).render()


class CacheTagMetaClass(type):
    """
    Metaclass used by CacheTag to save the Meta entries in a options field, and
    save the current class in the Node one.
    """

    def __new__(mcs, name, bases, attrs):
        klass = super(CacheTagMetaClass, mcs).__new__(mcs, name, bases, attrs)
        klass.options = klass._meta = klass.Meta()
        # One `Node` class for each `CacheTag` class, with a link on the reverse side too
        klass.Node = type('Node', (klass.Node, ), {'_cachetag_class_': klass})
        return klass


class CacheTag(object, metaclass=CacheTagMetaClass):
    """
    The main class of `django-adv-cache-tag` which does all the work.

    To change its behaviour, simply change one or more of these settings
    (see in the `Meta` class for details) :

        * ADV_CACHE_VERSIONING
        * ADV_CACHE_COMPRESS
        * ADV_CACHE_COMPRESS_LEVEL
        * ADV_CACHE_COMPRESS_SPACES
        * ADV_CACHE_INCLUDE_PK
        * ADV_CACHE_BACKEND
        * ADV_CACHE_VERSION
        * ADV_CACHE_RESOLVE_NAME

    Or inherit from this class and don't forget to register your tag :

        from adv_cache_tag.tag import CacheTag
        register = template.Library()
        class MyCacheTag(CacheTag):
            # change something
        MyCacheTag.register(register, 'my_cache')

    By inheriting you can change many things as CacheTag implements a lot of
    small methods
    """

    # Will change if the algorithm changes
    INTERNAL_VERSION = '1'
    # Used to separate internal version, template version, and the content
    VERSION_SEPARATOR = '::'

    # Regex used to reduce spaces/blanks (many spaces into one)
    RE_SPACELESS = re.compile(r'\s\s+')

    # generate a token for this site, based on the secret_key
    RAW_TOKEN = 'RAW_' + hashlib.sha1(
        b'RAW_TOKEN_SALT1' + force_bytes(hashlib.sha1(
            b'RAW_TOKEN_SALT2' + force_bytes(settings.SECRET_KEY)
        ).hexdigest())
    ).hexdigest()

    # tokens to use around the already parsed parts of the cached template
    RAW_TOKEN_START = template.BLOCK_TAG_START + RAW_TOKEN + template.BLOCK_TAG_END
    RAW_TOKEN_END = template.BLOCK_TAG_START + 'end' + RAW_TOKEN + template.BLOCK_TAG_END

    # internal use only: keep reference to templatetags functions
    _templatetags = {}
    # internal use only: name of the templatetags module to load for this class and subclasses
    _templatetags_modules = {}

    options = None
    Node = Node

    class Meta:
        """
        Options of this class. Accessible via cls.options or self.options.
        To force (and/or add) options in your own class, simply redefine a
        `Meta` class in your own main cache class with updated/add values
        """

        # If versioning is activated (internal versioning is always on)
        versioning = getattr(settings, 'ADV_CACHE_VERSIONING', False)

        # If the content will be compressed before caching
        compress = getattr(settings, 'ADV_CACHE_COMPRESS', False)
        compress_level = getattr(settings, 'ADV_CACHE_COMPRESS_LEVEL', zlib.Z_DEFAULT_COMPRESSION)

        # If many spaces/blanks will be converted into one
        compress_spaces = getattr(settings, 'ADV_CACHE_COMPRESS_SPACES', False)

        # If a "pk" (you can pass what you want) will be added to the cache key
        include_pk = getattr(settings, 'ADV_CACHE_INCLUDE_PK', False)

        # The cache backend to use (or use the "default" one)
        cache_backend = getattr(settings, 'ADV_CACHE_BACKEND', 'default')

        # Part of the INTERNAL_VERSION configurable via settings
        internal_version = getattr(settings, 'ADV_CACHE_VERSION', '')

        # If the fragment name should be resolved or taken as is
        resolve_fragment = getattr(settings, 'ADV_CACHE_RESOLVE_NAME', False)

    # Use a metaclass to use the right class in the Node class, and assign Meta to options

    def __init__(self, node, context):
        """
        Constructor of the Cache class:
            * preparing fields to be used later,
            * prepare the templatetag parameters
            * create the cache key
        """
        super(CacheTag, self).__init__()

        # the actual Node object
        self.node = node

        # the context used for the rendering
        self.context = context

        # indicate that we force regenerating the cache, even if it exists
        self.regenerate = bool(self.context.get('__regenerate__', False))

        # indicate if we only want html without parsing the nocache parts
        self.partial = bool(self.context.get('__partial__', False))

        # the content of the template, will be used through the whole process
        self.content = ''
        # the version used in the cached templatetag
        self.content_version = None

        # Final "INTERNAL_VERSION"
        if self.options.internal_version:
            self.INTERNAL_VERSION = force_bytes('%s|%s' % (self.__class__.INTERNAL_VERSION,
                                                           self.options.internal_version))
        else:
            self.INTERNAL_VERSION = force_bytes(self.__class__.INTERNAL_VERSION)

        self.VERSION_SEPARATOR = force_bytes(self.__class__.VERSION_SEPARATOR)

        # prepare all parameters passed to the templatetag
        self.expire_time = None
        self.version = None
        self.prepare_params()

        # get the cache and cache key
        self.cache = self.get_cache_object()
        self.cache_key = self.get_cache_key()

    def prepare_params(self):
        """
        Prepare the parameters passed to the templatetag
        """
        if self.options.resolve_fragment:
            self.fragment_name = self.node.fragment_name.resolve(self.context)
        else:
            self.fragment_name = str(self.node.fragment_name)
            # Remove quotes that surround the name
            for char in '\'\"':
                if self.fragment_name.startswith(char) or self.fragment_name.endswith(char):
                    if self.fragment_name.startswith(char) and self.fragment_name.endswith(char):
                        self.fragment_name = self.fragment_name[1:-1]
                        break
                    else:
                        raise ValueError('Number of quotes around the fragment name is incoherent')

        self.expire_time = self.get_expire_time()

        if self.options.versioning:
            self.version = force_bytes(self.get_version())

        self.vary_on = [template.Variable(var).resolve(self.context) for var in self.node.vary_on]

    def get_expire_time(self):
        """
        Return the expire time passed to the templatetag.
        Must be None or an integer.
        """
        try:
            expire_time = self.node.expire_time.resolve(self.context)
        except template.VariableDoesNotExist:
            raise template.TemplateSyntaxError('"%s" tag got an unknown variable: %r' %
                                               (self.node.nodename, self.node.expire_time.var))
        try:
            if expire_time is not None:
                expire_time = str(expire_time)
                if not expire_time.isdigit():
                    raise TypeError
                expire_time = int(expire_time)
        except (ValueError, TypeError):
            raise template.TemplateSyntaxError(
                '"%s" tag got a non-integer (or None) timeout value: %r' % (
                    self.node.nodename, expire_time
                )
            )

        return expire_time

    def get_version(self):
        """
        Return the stringified version passed to the templatetag.
        """
        if not self.node.version:
            return None
        try:
            version = smart_str('%s' % self.node.version.resolve(self.context))
        except template.VariableDoesNotExist:
            raise template.TemplateSyntaxError('"%s" tag got an unknown variable: %r' %
                                               (self.node.nodename, self.node.version.var))

        return '%s' % version

    def hash_args(self):
        """
        Take all the arguments passed after the fragment name and return a
        hashed version which will be used in the cache key
        """
        return hashlib.md5(force_bytes(':'.join([quote(force_bytes(var)) for var in self.vary_on]))).hexdigest()

    def get_pk(self):
        """
        Return the pk to use in the cache key. It's the first version of the
        templatetag arguments after the fragment name
        """
        return self.vary_on[0]

    def get_base_cache_key(self):
        """
        Return a string with format placeholder used as a source to compute the
        final cache key.
        Placeholders are :
            * %(nodename)s : the name of the templatetag
            * %(name)s : the fragment name passed to the templatetag
            * %(pk)s : the return of the `get_pk` method, passed only if `include_pk` is True
            * %(hash)s : the return of the `hash_args` method
        """
        if self.options.include_pk:
            return 'template.%(nodename)s.%(name)s.%(pk)s.%(hash)s'
        else:
            return 'template.%(nodename)s.%(name)s.%(hash)s'

    def get_cache_key_args(self):
        """
        Return the arguments to be passed to the base cache key returned by `get_base_cache_key`.
        """
        cache_key_args = dict(
            nodename=self.node.nodename,
            name=self.fragment_name,
            hash=self.hash_args(),
        )
        if self.options.include_pk:
            cache_key_args['pk'] = self.get_pk()

        return cache_key_args

    def get_cache_key(self):
        """
        Compute and return the final cache key, using return values of
        `get_base_cache_key` and `get_cache_key_args`.
        """
        return self.get_base_cache_key() % self.get_cache_key_args()

    def get_cache_object(self):
        """
        Return the cache object to be used to set and get the values in cache.
        By default it's the default cache defined by django, but it can be
        every object with a `get` and a `set` method (or not, if `cache_get`
        and `cache_set` methods are overridden)
        """
        return get_cache(self.node.cache_backend or self.options.cache_backend)

    def cache_get(self):
        """
        Get content from the cache
        """
        return self.cache.get(self.cache_key)

    def cache_set(self, to_cache):
        """
        Set content into the cache
        """
        self.cache.set(self.cache_key, to_cache, self.expire_time)

    def join_content_version(self, to_cache):
        """
        Add the version(s) to the content to cache : internal version at first
        and then the template version if versioning is activated.
        Each version, and the content, are separated with `VERSION_SEPARATOR`.
        This method is called after the encoding (if "compress" or
        "compress_spaces" options are on)
        """
        parts = [self.INTERNAL_VERSION]
        if self.options.versioning:
            parts.append(force_bytes(self.version))
        parts.append(force_bytes(to_cache))

        return self.VERSION_SEPARATOR.join(parts)

    def split_content_version(self):
        """
        Remove and return the version(s) from the cached content. First the
        internal version, and if versioning is activated, the template one.
        And finally save the content, but only if all versions match.
        The content saved is the encoded one (if "compress" or
        "compress_spaces" options are on). By doing so, we avoid decoding if
        the versions didn't match, to save some cpu cycles.
        """
        try:
            nb_parts = 2
            if self.options.versioning:
                nb_parts = 3

            parts = self.content.split(self.VERSION_SEPARATOR, nb_parts - 1)
            assert len(parts) == nb_parts

            self.content_internal_version = parts[0]
            if self.options.versioning:
                self.content_version = parts[1]

            self.content = parts[-1]
        except Exception:
            self.content = None

    def decode_content(self):
        """
        Decode (decompress...) the content got from the cache, to the final
        html
        """
        self.content = pickle.loads(zlib.decompress(self.content))

    def encode_content(self):
        """
        Encode (compress...) the html to the data to be cached
        """
        return zlib.compress(pickle.dumps(self.content), self.options.compress_level)

    def render_node(self):
        """
        Render the template and save the generated content
        """
        self.content = self.node.nodelist.render(self.context)

    def create_content(self):
        """
        Render the template, apply options on it, and save it to the cache.
        """
        self.render_node()

        if self.options.compress_spaces:
            self.content = self.RE_SPACELESS.sub(' ', self.content)

        if self.options.compress:
            to_cache = self.encode_content()
        else:
            to_cache = self.content

        to_cache = self.join_content_version(to_cache)

        try:
            self.cache_set(to_cache)
        except Exception:
            if is_template_debug_activated():
                raise
            logger.exception('Error when saving the cached template fragment')

    def load_content(self):
        """
        It's the main method of the class.
        Try to load the template from cache, get the versions and decode the
        content.
        If something was wrong during this process (or if we had a
        `__regenerate__` value to True in the context), create new content and
        save it in cache.
        """

        self.content = None

        if not self.regenerate:
            try:
                self.content = self.cache_get()
            except Exception:
                if is_template_debug_activated():
                    raise
                logger.exception('Error when getting the cached template fragment')

        try:

            assert self.content

            self.split_content_version()

            assert self.content

            if self.content_internal_version != self.INTERNAL_VERSION or (
                    self.options.versioning and self.content_version != self.version):
                self.content = None

            assert self.content

            if self.options.compress:
                self.decode_content()

        except Exception:
            self.create_content()

        self.content = smart_str(self.content)


    def render(self):
        """
        Try to load content (from cache or by rendering the template).
        If it fails, return an empty string or raise the exception if it's a
        TemplateSyntaxError.
        With this, we can no parse and render the content included in the
        {% nocache %} blocks, but only if we have have this tag and if we don't
        have `__partial__` to True in the context (in this case we simple
        return the html with the {% nocache %} block not parsed.
        """
        try:
            self.load_content()
        except template.TemplateSyntaxError:
            raise
        except Exception:
            if is_template_debug_activated():
                raise
            logger.exception('Error when rendering template fragment')
            return ''

        if self.partial or self.RAW_TOKEN_START not in self.content:
            return self.content

        return self.render_nocache()

    @staticmethod
    def get_all_tags_and_filters_by_function():
        """
        Return a dict with all the template tags (in the `tags` entry) and filters (in the
        `filters` entry) that are available.
        Both entries are a dict with the function as key, and a tuple with (library name, function
        name) as value.
        This is cached after the first call.
        """

        libraries = get_template_libraries()

        force = False

        # We'll force the update of the cache if new libraries where added
        if hasattr(CacheTag.get_all_tags_and_filters_by_function, '_len_libraries'):
            if len(libraries) != CacheTag.get_all_tags_and_filters_by_function._len_libraries:
                force = True

        if force or not hasattr(CacheTag.get_all_tags_and_filters_by_function, '_cache'):
            CacheTag.get_all_tags_and_filters_by_function._len_libraries = len(libraries)
            available_tags = {}
            available_filters = {}

            for lib_name, lib in libraries.items():
                available_tags.update(
                    (function, (lib_name, tag_name))
                    for tag_name, function
                    in lib.tags.items()
                )
                available_filters.update(
                    (function, (lib_name, filter_name))
                    for filter_name, function
                    in lib.filters.items()
                )

            CacheTag.get_all_tags_and_filters_by_function._cache = {
                'tags': available_tags,
                'filters': available_filters
            }

        return CacheTag.get_all_tags_and_filters_by_function._cache

    @classmethod
    def get_templatetag_module(cls):
        """
        Return the templatetags module name for which the current class is used.
        It's used to render the nocache blocks by loading the correct module
        """
        if cls not in CacheTag._templatetags_modules:
            # find the library including the main templatetag of the current class
            all_tags = cls.get_all_tags_and_filters_by_function()['tags']
            CacheTag._templatetags_modules[cls] = all_tags[CacheTag._templatetags[cls]['cache']][0]
        return CacheTag._templatetags_modules[cls]

    def render_nocache(self):
        """
        Render the `nocache` blocks of the content and return the whole
        html
        """
        tmpl = template.Template(''.join([
            # start by loading the cache library
            template.BLOCK_TAG_START,
            'load %s' % self.get_templatetag_module(),
            template.BLOCK_TAG_END,
            # and surround the cached template by "raw" tags
            self.RAW_TOKEN_START,
            self.content,
            self.RAW_TOKEN_END,
        ]))
        return tmpl.render(self.context)

    @classmethod
    def get_template_node_arguments(cls, tokens):
        """
        Return the arguments taken from the templatetag that will be used to the
        Node class.
        Take a list of all tokens and return a list of real tokens. Here
        should be done some validations (number of tokens...) and eventually
        some parsing...
        """
        if len(tokens) < 3:
            raise template.TemplateSyntaxError(
                "'%r' tag requires at least 2 arguments." % tokens[0])
        return tokens[1], tokens[2], tokens[3:]

    @classmethod
    def register(cls, library_register, nodename='cache', nocache_nodename='nocache'):
        """
        Register all needed templatetags, with these parameters :
            * library_register : the `register` object (result of
                `template.Library()`) in your templatetag module
            * nodename : the node to use for the cache templatetag (the default
                is "cache")
            * nocache_nodename : the node to use for the nocache templatetag
        """

        if cls in CacheTag._templatetags:
            raise RuntimeError('The adv-cache-tag class %s is already registered' % cls)

        CacheTag._templatetags[cls] = {}

        def templatetag_cache(parser, token):
            """
            Return a new Node object for the main cache templatetag
            """
            nodelist = parser.parse(('end%s' % nodename,))
            parser.delete_first_token()
            args = cls.get_template_node_arguments(token.contents.split())
            return cls.Node(nodename, nodelist, *args)

        library_register.tag(nodename, templatetag_cache)
        CacheTag._templatetags[cls]['cache'] = templatetag_cache

        def templatetag_raw(parser, token):
            """
            Return a TextNode with all html not parsed, used for templatetags
            that need to not be parsed : the `nocache` one and the `RAW` one,
            used to surround cached html (to be not parsed again)
            Based on http://www.holovaty.com/writing/django-two-phased-rendering/
            """

            # Whatever is between {% nocache %} and {% endnocache %} will be preserved as
            # raw, un-rendered template code.

            text = []
            parse_until = 'end%s' % token.contents
            tag_mapping = {
                TOKEN_TEXT: ('', ''),
                TOKEN_VAR: ('{{', '}}'),
                TOKEN_BLOCK: ('{%', '%}'),
                TOKEN_COMMENT: ('{#', '#}'),
            }
            # By the time this template tag is called, the template system has already
            # lexed the template into tokens. Here, we loop over the tokens until
            # {% endraw %} and parse them to TextNodes. We have to add the start and
            # end bits (e.g. "{{" for variables) because those have already been
            # stripped off in a previous part of the template-parsing process.
            while parser.tokens:
                token = parser.next_token()
                if token.token_type == TOKEN_BLOCK and token.contents == parse_until:
                    return template.TextNode(''.join(text))
                start, end = tag_mapping[token.token_type]
                text.append('%s%s%s' % (start, token.contents, end))
            parser.unclosed_block_tag(parse_until)

        library_register.tag(cls.RAW_TOKEN, templatetag_raw)
        CacheTag._templatetags[cls]['raw'] = templatetag_raw

        def templatetag_nocache(parser, token):
            """
            Return a TextNode with raw html from the `nocache` templatetag,
            and surround it with `endRAW` and `RAW` (precisely
            `cls.RAW_TOKEN_END` and `cls.RAW_TOKEN_START`).
            So for
                {% nocache %}foo{% endnocache %}
            we get
                {% endRAW... %}foo{% RAW... %}
            When the main cache templatetag content will be loaded from cache,
            it will be surrounded by the same templatetags, reversed.
            So if at first we had
                {% cache %}bar{% nocache %}foo{% endnocache %}baz{% endcache %}
            The cached version will be
                bar{% endRAW... %}foo{% RAW... %}baz
            And the final html to be rendered will be
                {% RAW... %}bar{% endRAW... %}foo{% RAW... %}baz{% endRAW... %}
            And the html within `RAW` and `endRAW` will not be parsed, as wanted
            """

            # We'll load in the no-cache part all template tags and filters loaded in the main
            # template, to be able to use it when the no-cache will be rendered

            all_tags_and_filters = cls.get_all_tags_and_filters_by_function()
            available_tags = all_tags_and_filters['tags']
            available_filters = all_tags_and_filters['filters']

            needed = {}
            current_module = cls.get_templatetag_module()

            for function in parser.tags.values():
                if function in available_tags:
                    lib, name = available_tags[function]
                    if lib == current_module:
                        continue
                    needed.setdefault(lib, set()).add(name)

            for function in parser.filters.values():
                if function in available_filters:
                    lib, name = available_filters[function]
                    if lib == current_module:
                        continue
                    needed.setdefault(lib, set()).add(name)

            load_string = ''.join(
                '%sload %s from %s%s' % (
                    template.BLOCK_TAG_START,
                    ' '.join(names),
                    lib,
                    template.BLOCK_TAG_END,
                )
                for lib, names in needed.items()
            )

            node = templatetag_raw(parser, token)
            node.s = cls.RAW_TOKEN_END + load_string + node.s + cls.RAW_TOKEN_START
            return node

        library_register.tag(nocache_nodename, templatetag_nocache)
        CacheTag._templatetags['nocache'] = templatetag_nocache



from django import VERSION
from django.test import TestCase


if VERSION < (1, 7):
    from adv_cache_tag.compat import get_cache

    class TestCase(TestCase):
        def setUp(self):
            super(TestCase, self).setUp()

            # Override default cache in django < 1.7 because it is initialized before our
            # `override_settings`
            from django.core import cache
            cache.cache = get_cache('default')



import os.path

DEBUG = True
TEMPLATE_DEBUG = DEBUG

SECRET_KEY = 'm-92)2et+&&m5f&#jld7-_1qanq*n9!z90xc@+wx6y8d6y#w6t'

BASE_DIR = os.path.dirname(__file__)

def absolute_path(path):
    return os.path.normpath(os.path.join(BASE_DIR, path))

SITE_ID = 1
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': absolute_path('database.sqlite3'),
    }
}

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'adv_cache_tag',
    'adv_cache_tag.tests.testproject.adv_cache_test_app',
]

MIDDLEWARE_CLASSES = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'debug': False
        }
    },
]




try:
    from django.core.cache import caches
except ImportError:
    # Django < 1.7
    from django.core.cache import get_cache
else:
    def get_cache(name):
        return caches[name]

try:
    from django.template import BLOCK_TAG_START
except ImportError:
    # Django >= 1.8
    from django.template import base as template
else:
    from django import template


def get_template_libraries():
    try:
        from django.template.base import libraries
    except ImportError:
        # Django >= 1.9
        from django.template import Engine
        libraries = Engine.get_default().template_libraries

    return libraries



#!/usr/bin/env python

import sys

from setuptools import setup



if sys.version_info < (3,):
    sys.stderr.write('Fatal error: django-adv-cache-tag version >1 only works with python 3. '
                     'Use version <1.0 for python 2.\n')
    sys.exit(1)

setup()
